/*
   prevod z WAV: ffmpeg -f s16le -acodec pcm_s16le -i input.wav output.pcm
   ffmpeg -i file.wav -f s16le -ar 48k -acodec pcm_s16le file.raw

   prevod do WAV: ffmpeg -f s16le -ar 48k -ac 2 -i input.pcm output.wav
*/

#define _GNU_SOURCE
#include "stdio.h"
#include "sys/types.h"
#include "unistd.h"
#include "string.h"
#include "stdlib.h"
#include "signal.h"
#include "unistd.h"
#include "fcntl.h"
#include "sys/time.h"

#include <e-hal.h>


#define BUFLEN 6000 // velikost docasnych promennych - odesilani, prijem

#define DEBUG 1

typedef struct {
  int frame_size;
  int sample_rate;
  int channels;
  int application;
  int bitrate;
  int max_frame_size;
  int max_packet_size;
} parameter_list;

typedef struct {
  parameter_list params;
  int file_id;
  int chunk_id;
} encoder_parameters;

typedef struct {
  int opened;
  long filesize;
  char *name;
  char *output_name;
  struct timeval start_time;
  struct timeval end_time;
} fileinfo;

typedef struct {
  int col;
  int row;
} proc_loc;

char *basename(char const *path);

void mainloop();

int create_encoders(int number);
int create_decoders(int number);

int row(int process);
int col(int process);

int encode_frame(int16_t *input_frame, encoder_parameters encoder_params);
int decode_frame(unsigned char *input_frame, int input_size, encoder_parameters encoder_params);

int load_file(const char *filename, char **output);
int create_output_buffer(int buffer_size, char **buffer);
int save_buffer(const char *filename, char *buffer, int buffer_size);
int read_chunk(char *ramfile, int offset, int chunk_size, int16_t *output);
int save_chunk(char *buffer, int offset, int chunk_size, int16_t *input);

void get_core_mem(int num_process);

int process2core(int process);
int core2process(int core);
void set_core_status(int core, int status);
void set_process_status(int process, int status);

void show_progress();

int *dev_mem;
//int *core_mem;
e_mem_t emem;

//int ready_to_process=2;

parameter_list params;
encoder_parameters encoder_params;

char buffer[BUFLEN];

encoder_parameters recv_params;
int recv_size;

   char **fin;
   char **fout;
   int *chunk;
   fileinfo *file_info;
   int file_id;
   int num_files;
   int *process_ready;

   e_epiphany_t *dev;



int main(int argc, char **argv)
{
if (DEBUG==1) {printf("\n################ opus server ################\n\n");}
  // nastaveni parametru enkoderu

    params.frame_size = 1920;
    params.sample_rate = 48000;
    params.channels=2;
    params.application=2049;  // konstanta pro OPUS_APPLICATION_AUDIO
    params.bitrate=64000;
    params.max_frame_size=1920;
    params.max_packet_size=params.max_frame_size*params.channels;

// inicializace epiphany
e_init(NULL);
e_reset_system();
e_alloc(&emem,0x01000000,sizeof(int)*16);
    

num_files=(argc-1)/2;


  // soubory

fin=(char**)malloc(sizeof(char*)*num_files);
fout=(char**)malloc(sizeof(char*)*num_files);

chunk=(int*)malloc(sizeof(int)*num_files);
process_ready=(int*)malloc(sizeof(int)*num_files*2);

file_info=(fileinfo*)malloc(sizeof(fileinfo)*num_files);
dev=(e_epiphany_t*)malloc(sizeof(e_epiphany_t)*num_files*2);

//core_mem=(int*)malloc(sizeof(int)*16);   //offsety na jadrech
dev_mem=(int*)malloc(sizeof(int)*num_files*2); //offsety v procesech

if ((argc<3) || (!(argc % 2))) {printf("Usage: opus_server input1 output1 input2 output2 ... \n");exit(0);}

int i;int y=0;
for (i=1;i<argc;i++)
{
if (DEBUG==1) {printf("Input %d: %s\nOutput %d: %s\n",y,argv[i],y,argv[i+1]); }

file_info[y].filesize=load_file(argv[i],&fin[y]);
create_output_buffer(file_info[y].filesize,&fout[y]);

file_info[y].opened=1;
file_info[y].name=basename(argv[i]);
file_info[y].output_name=basename(argv[i+1]);

process_ready[i]=0; // pomocna promenna pro zjisteni jestli jsou spustene vsechny enkodery a dekodery
i++;
y++;
}

create_encoders(num_files);
create_decoders(num_files);
get_core_mem(num_files);

int testloop=10;

while(testloop>0)
{
  mainloop();
//  testloop--;
}

return 0;
}


void get_core_mem(int num_process)
{   if(DEBUG==1) {printf("running processes: %d\n",num_process*2);}
    printf("getting offsets from cores.. \n");
    int core_mem[16];
    e_read(&emem,0,0,0x0,core_mem,sizeof(int)*16);

    int i;int y=0;
    for(i=0;i<num_process;i++)
    {
      if (DEBUG==1){printf("Encoder %d - int_buffer offset: 0x%04x\n",i,core_mem[i]);}
      dev_mem[y++]=core_mem[i];
      core_mem[i]=1;
    }
    for(i=0;i<num_process;i++)
    {
      if (DEBUG==1){printf("Decoder %d - int_buffer offset: 0x%04x\n",i,core_mem[i+8]);}
      dev_mem[y++]=core_mem[i+8];
      core_mem[i+8]=1;
    }
    e_write(&emem,0,0,0x0,core_mem,sizeof(int)*16);
}


void mainloop()
{
    // precteni prijateho paketu a dekodovani
  int z;
  for(z=0;z<(num_files*2);z++)
  {
    int tmp=get_process_status(z);
    //printf("process #%d: status %d\n",z,tmp);
    if (tmp==1) // prisly data pro zpracovani
    {
    
    if(DEBUG==1) {printf("checking data on process #%d (core #%d, offset 0x%04x)..",z,process2core(z),dev_mem[z]);}
    e_read(&dev[z],0,0,dev_mem[z],buffer,BUFLEN);

    switch (buffer[0])
    {
      case 'D':{// prisel dekodovany frame - ulozit
                if (DEBUG==1) {printf("server    - Received decoded frame.\n");}
                // ulozeni dat z paketu do promennych
                memcpy(&recv_params, buffer+sizeof(char), sizeof(encoder_parameters));
                memcpy(&recv_size, buffer+sizeof(char)+sizeof(encoder_parameters), sizeof(int));
                int16_t recv_data[recv_size*recv_params.params.channels];
                memcpy(&recv_data, buffer+sizeof(char)+sizeof(encoder_parameters)+sizeof(int), recv_size*recv_params.params.channels*sizeof(int16_t));
                // ulozeni dekodovaneho ramce

                //save_frame(fout[recv_params.file_id], recv_params.params.channels, recv_size, (int16_t*)&recv_data);
                
                int writed;
                writed=save_chunk(fout[recv_params.file_id], chunk[recv_params.file_id], (sizeof(short)*recv_params.params.channels*recv_size), (int16_t*)&recv_data);
                if (DEBUG==1) {printf("chunk writed: %d bytes\n",writed);}

                recv_params.chunk_id=++chunk[recv_params.file_id];
                if (DEBUG==0) {show_progress();}

                // nacteni dalsiho ramce ze souboru a odeslani enkoderu
                if (((chunk[recv_params.file_id])*(sizeof(short)*recv_params.params.channels*recv_params.params.frame_size))>file_info[recv_params.file_id].filesize)
                {
                  if (DEBUG==1) {printf("server    - File end - %s\n",file_info[recv_params.file_id].name);}
                  gettimeofday(&(file_info[recv_params.file_id].end_time), NULL);
                  file_info[recv_params.file_id].opened=0;

                  int saved;
                  saved=save_buffer(file_info[recv_params.file_id].output_name, fout[recv_params.file_id], file_info[recv_params.file_id].filesize);
                  //zatim pocitam s velikosti identickou, bude lepsi vypocitat velikost - pocty chunku * recv_size
                  if (DEBUG==1) {printf("file saved: %d bytes\n",saved);}

                  chunk[recv_params.file_id]++;
                  if (DEBUG==0) {show_progress();}

                  int i;int done=1;
                  for (i=0;i<num_files;i++) {if (file_info[i].opened==1) {done=0;}}
                  if (done==1) {
                  exit(0);
                  //jeste killnout procesy
                  }
                }
                else
                {
                  int readed;
                  int16_t loaded_data[recv_params.params.channels*recv_params.params.frame_size];
		  int diff=(chunk[recv_params.file_id]+1)*(sizeof(short)*recv_params.params.channels*recv_params.params.frame_size)-file_info[recv_params.file_id].filesize;
                  if (diff>0)
                  {  //posledni chunk ze souboru
                    readed=read_chunk(fin[recv_params.file_id], chunk[recv_params.file_id], (recv_params.params.channels*recv_params.params.frame_size)-diff, loaded_data);
                  }
                  else
                  {
                    readed=read_chunk(fin[recv_params.file_id], chunk[recv_params.file_id], (recv_params.params.channels*recv_params.params.frame_size), loaded_data);
                  }
                  if (DEBUG==1) {printf("chunk readed: %d bytes\n",readed);}
                  //load_frame(fin[recv_params.file_id], recv_params.params.channels, recv_params.params.frame_size, in)
                  encode_frame(loaded_data,recv_params);
                }

      }break;

      case 'E':{// prisel zakodovany frame - dekodovat
                if (DEBUG==1) {printf("server    - received encoded frame.\n");}
                // ulozeni dat z paketu do promennych
                memcpy(&recv_params, buffer+sizeof(char), sizeof(encoder_parameters));
                memcpy(&recv_size, buffer+sizeof(char)+sizeof(encoder_parameters), sizeof(int));
                unsigned char recv_data[recv_size];
                memcpy(&recv_data, buffer+sizeof(char)+sizeof(encoder_parameters)+sizeof(int), recv_size*sizeof(unsigned char));
                // odeslani zakodovanych dat dekoderu
                decode_frame(recv_data,recv_size,recv_params);
		//set_process_status(recv_params.file_id+num_files,2); // 2 - jsou tam poslane data na zpracovani
      }break;

      case 'S':{// prisly informace o stavu - spusteni enkoderu/dekoderu
                if ((buffer[1])=='D')
                { // ohlasil se dekoder
                  if (DEBUG==1) {printf("Decoder %d ready.\n",buffer[2]);}
                  process_ready[num_files+buffer[2]]=1;
		  set_process_status(num_files+buffer[2],0); // 0 - cekej
                }
                else if ((buffer[1])=='E')
                { // ohlasil se enkoder
                  if (DEBUG==1) {printf("Encoder %d ready.\n",buffer[2]);}
                  process_ready[buffer[2]]=1;
		  set_process_status(buffer[2],0); // 0 - cekej
                }

                int y;int ready=1;
                for (y=0;y<(num_files*2);y++) {if (process_ready[y]==0) {ready=0;}}
                if (ready==1) // nacteni prvniho framu ze souboru a odeslani enkoderu ktery se ohlasil jako ready
                {
                 if (DEBUG==1) {printf("All decoders and encoders ready. Files:%d Starting...\n",num_files);}
                 for (y=0;y<(num_files);y++)
                 {int16_t loaded_data[params.channels*params.frame_size];
                  int readed;
		  readed=read_chunk(fin[y], 0, (params.channels*params.frame_size), loaded_data);
                  if (readed>0)
                  {
                    if (DEBUG==1) {printf("chunk readed: %d bytes\n",readed);}
                    encoder_params.chunk_id=0;
                    encoder_params.file_id=y;
                    encoder_params.params=params;
                    gettimeofday(&(file_info[y].start_time), NULL);
                    if (DEBUG==0) {show_progress(file_info[(int)buffer[2]],0);}
                    if (DEBUG==1) {printf("file %d - encoding started. \n",y);}
		    encode_frame(loaded_data,encoder_params);
		    //set_process_status(y,2); // 2 = jsou tam poslane data na zpracovani
                  }
		  else {printf("Error reading frame! - file %d\n",y);}
                 }
                }
      }break;

      default: // neznamy paket
                printf("Unknown data!\n");
                //printf("Data: %.*s\n",30,buffer);
                if (DEBUG==1) {printf("RAW data: %02x %02x %02x %02x %02x %02x %02x %02x\n",buffer[0],buffer[1],buffer[2],buffer[3],buffer[4],buffer[5],buffer[6],buffer[7]);}
               // mohlo by se vypsat napr. prvnich 16 bajtu v cyklu, ale neni nutne zatim.
      break;
    }
    }

  }

}

int create_encoders(int number)
{ //
  if(DEBUG==1) {printf("starting encoders... \n");}
  int y;
  for (y=0;y<number;y++)
    {
      // enkodery - jadra 0-7
      if (DEBUG==1) {printf("opening encoder #%d (process #%d, core #%d) on [%d,%d]... ",y,y,y,row(y),col(y));}
      e_open(&dev[y],row(y),col(y),1,1);
      e_reset_group(&dev[y]);
      e_load("opus_encode.srec",&dev[y],0,0,E_TRUE);
      if (DEBUG==1) {printf("started.\n");}
    }
return 1;
}

int create_decoders(int number)
{ // stejne jako create_encoders, ale s dekodery
  if(DEBUG==1) {printf("starting decoders... \n");}
  int y;
  for (y=0;y<number;y++)
    {
      // dekodery - jadra 8-15
      if (DEBUG==1) {printf("opening decoder #%d (process #%d, core #%d) on [%d,%d]... ",y,core2process(y+8),y+8,row(y+8),col(y+8));}
      e_open(&dev[y+number],row(y+8),col(y+8),1,1);
      e_reset_group(&dev[y+number]);
      e_load("opus_decode.srec",&dev[y+number],0,0,E_TRUE);
      if (DEBUG==1) {printf("started.\n");}
    }
return 1;
}


int encode_frame(int16_t *input_frame, encoder_parameters encoder_params)
{ // poslani framu enkoderu - priprava paketu, odeslani
  // vlastni struktura paketu - nejprve parametry (typ encoder_parameters), pak9;10M samotne data.
  
  char output_data[(encoder_params.params.channels*encoder_params.params.frame_size*sizeof(int16_t))+sizeof(encoder_parameters)];
  memcpy(output_data,&encoder_params,sizeof(encoder_parameters));
  memcpy(output_data+sizeof(encoder_parameters),input_frame,encoder_params.params.frame_size*encoder_params.params.channels*sizeof(int16_t));
  if (DEBUG==1) {
  printf("server    - chunk #%d - data sent to encoder #%d\n", encoder_params.chunk_id, encoder_params.file_id);
  }
  //sendto(socketDescriptor_send, output_data, sizeof(encoder_parameters)+(encoder_params.params.channels*encoder_params.params.frame_size*sizeof(int16_t)), 0, (struct sockaddr *) &serverAddress_send, sizeof(serverAddress_send));
  //e_write(&dev[encoder_params.file_id],0,0,dev_mem[encoder_params.file_id],output_data,sizeof(output_data));
  set_process_status(encoder_params.file_id,2); 
  return 1;
}

int decode_frame(unsigned char *input_frame, int input_size, encoder_parameters encoder_params)
{ // stejne jako encode_frame, akorat jina struktura paketu a jiny prijemce

  char output_data[(input_size*sizeof(unsigned char))+(sizeof(int))+(sizeof(encoder_parameters))];
  memcpy(output_data,&encoder_params,sizeof(encoder_parameters));
  memcpy(output_data+sizeof(encoder_parameters),&input_size,sizeof(int));
  memcpy(output_data+sizeof(encoder_parameters)+sizeof(int),input_frame,input_size*sizeof(unsigned char));
  if (DEBUG==1) {printf("server    - chunk #%d - data sent to decoder %d\n", encoder_params.chunk_id, encoder_params.file_id);}
  //sendto(socketDescriptor_send, output_data, sizeof(int)+sizeof(encoder_parameters)+input_size*sizeof(unsigned char), 0, (struct sockaddr *) &serverAddress_send, sizeof(serverAddress_send));
  //e_write(&dev[encoder_params.file_id+num_files],0,0,dev_mem[encoder_params.file_id+num_files],output_data,sizeof(output_data));
  set_process_status(encoder_params.file_id,2); 
  return 1;
}

int load_file(const char *filename, char **output)
{ // nacteni souboru do RAM

    int filesize = 0;
	FILE *f = fopen(filename, "r");
	if (f == NULL) { return 0; }

	fseek(f, 0, SEEK_END);
	filesize = ftell(f);
	fseek(f, 0, SEEK_SET);

	
        *output = (char*)malloc(sizeof(char)*filesize+1);
        
	//memmap?
	if (filesize != fread(*output, sizeof(char), filesize, f))  { free(*output); return 0; }
	fclose(f);
	//*output[filesize] = '\0';

	
	return filesize;
}

int create_output_buffer(int buffer_size, char **buffer)
{ // vytvoreni output bufferu v RAM
    if (buffer_size>0)
    {
      buffer_size=buffer_size*1.1; // nejaka rezerva, vysledny soubor nemusi byt 1:1
      *buffer = (char*)malloc(sizeof(char)*buffer_size+1);
	  return buffer_size;
	}
	else {return 0;}
}

int save_buffer(const char *filename, char *buffer, int buffer_size)
{ // ulozeni output bufferu do souboru

	FILE *f = fopen(filename, "w");
	if (f == NULL) { return 0; }
	
	if (buffer_size != fwrite(buffer, sizeof(char), buffer_size, f))  { return 0; }

	fclose(f);
	return buffer_size;
}

int read_chunk(char *ramfile, int offset, int chunk_size, int16_t *output)
{ // nacteni chunku z RAM - souboru
  int i;
  // nacte a rovnou prevede do int16
  for (i=0;i<chunk_size;i++)
  {
    output[i]=ramfile[(offset*chunk_size)+(2*i+1)]<<8|ramfile[(offset*chunk_size)+(2*i)];
    //if (output[i]=='\0') {printf("reading-EOF (position %d)\n",i);break;}     // blbost
  }
  return i;
//osetrit konec souboru?
}

int save_chunk(char *buffer, int offset, int chunk_size, int16_t *input)
{ // ulozeni chunku do bufferu
  // vezme int16 a rovnou ulozi jako char
  int i;
  for (i=0;i<chunk_size;i++)
  {
    buffer[(offset*chunk_size)+(2*i)]=input[i]&0xFF;
    buffer[(offset*chunk_size)+(2*i+1)]=(input[i]>>8)&0xFF;
  }
  return i;
}




char *basename(char const *path)
{
    char *s = strrchr(path, '/');
    if (!s) {return strdup(path);} else {return strdup(s + 1);}
}

void show_progress()
{
  int progress;
  int i;int y;
  int done=1;

  for(i=0;i<num_files;i++)
  {
    printf("%s \t\t[",file_info[i].name);
    progress=100*(chunk[i]*params.channels*params.frame_size*sizeof(short))/(file_info[i].filesize);
    for(y=0;y<(progress/5);y++){printf("#");}
    for(y=0;y<(20-(progress/5));y++){printf(" ");}
    printf("] %d%% ",progress);
    if (file_info[i].end_time.tv_usec>0) {printf("\ttime: %fs",(file_info[i].end_time.tv_sec + (1.0/1000000) * file_info[i].end_time.tv_usec)-(file_info[i].start_time.tv_sec + (1.0/1000000) * file_info[i].start_time.tv_usec));} else  {done=0;}
    printf("\n");
  }

 if (done==0) { for(i=0;i<num_files;i++) {printf("%c[A", 27);printf("%c[2K", 27);}}

}

// ktery proces patri na jake jadro

int row(int process)
{
switch(process)
   {
    case 0: case 1: case 2: case 3: return 0;
    case 4: case 5: case 6: case 7: return 1;
    case 8: case 9: case 10: case 11: return 2;
    case 12: case 13: case 14: case 15: return 3;
    default:break;
   }
return -1;
}


int col(int process)
{
switch(process)
   {
    case 0: case 4: case 8: case 12: return 0;
    case 1: case 5: case 9: case 13: return 1;
    case 2: case 6: case 10: case 14: return 2;
    case 3: case 7: case 11: case 15: return 3;
    default:break;
   }
return -1;
}

int process2core(int process)
{if (process>=num_files) {return(process-num_files+8);} else {return process;}}

int core2process(int core)
{if (core<8) {return core;} else {return core-8+num_files;}}

void set_process_status(int process, int status)
{set_core_status(process2core(process),status);}

int get_process_status(int process)
{get_core_status(process2core(process));}

int get_core_status(int core)
{
  int tmp[16];
  e_read(&emem,0,0,0x0,tmp,sizeof(int)*16);
  return tmp[core];
}

void set_core_status(int core, int status)
{
// int ret;
  int tmp[16];
  e_read(&emem,0,0,0x0,tmp,sizeof(int)*16);
  tmp[core]=status;
  e_write(&emem,0,0,0x0,tmp,sizeof(int)*16); 
//printf("write: %db;",ret);
}



